// Copyright (C) 2002 Takeshi Morimoto <morimoto@takopen.cs.uec.ac.jp>
// All rights reserved.
package yab.agent;

import java.net.*;
import java.util.Random;
import yab.agent.*;
import yab.agent.object.*;
import yab.io.*;

public abstract class Agent extends Thread implements Constants {
    private final RCRSSProtocolSocket m_socket;
    protected final DisasterSpace world;
    protected final Random random;

    protected Agent(int agentType, InetAddress kernelAddress, int kernelPort) {
        m_socket = new RCRSSProtocolSocket(kernelAddress, kernelPort);
        socket().akConnect(0, agentType);
        Object data = socket().receive();
        if (data instanceof KaConnectError)
            throw new Error(((KaConnectError) data).reason);
        KaConnectOk ok = (KaConnectOk) data;
        world = new DisasterSpace(ok.selfId, ok.selfAndMap);
        socket().akAcknowledge(self().id);

        System.out.println(self());
        random = new Random((long) self().id);
    }

    /** This method decides action of the agent at each cycle. */
    protected abstract void act() throws ActionCommandException;

    /** This method decides reaction to hearing <code>message</code>
     *  from <code>sender</code>.
     */
    protected void hear(RealObject sender, String message) { }
    protected void hearSay(RealObject sender, String message) { hear(sender, message); }
    protected void hearTell(RealObject sender, String message) { hear(sender, message); }

    protected void prepareForAct() { /* do nothing */ }
    protected int utteranceLimit() { return UTTERANCE_LIMIT; }
    protected abstract int hearingLimit();

    protected RCRSSProtocolSocket socket() { return m_socket; }
    protected int time() { return world.time(); }
    private RealObject self() { return world.self; }

    protected final Property distancePrp = new Property() {
            public Object eval(Object obj) {
                return new Integer(self().distance((RealObject) obj)); }};

    private int m_numUtterance;
    private int m_numHearing;

    public void run() {
        while (true) {
            Object data = socket().receive();
            if (data instanceof KaSense) {
                KaSense sense = (KaSense) data;
                world.update(sense.selfAndMap, sense.time);
                m_numUtterance = 0;
                m_numHearing = 0;
                prepareForAct();
                try {
                    act();
                } catch (ActionCommandException sce) { /* do nothing */ }
            } else if (data instanceof KaHear) {
                KaHear hear = (KaHear) data;
                if (hear.type != ProtocolConstants.KA_HEAR) {
                    RealObject sender = world.get(hear.senderId);
                    if (sender != self() && m_numHearing ++ < hearingLimit()) {
                        switch (hear.type) {
                        case ProtocolConstants.KA_HEAR_SAY:
                            hearSay(sender, hear.message);
                            break;
                        case ProtocolConstants.KA_HEAR_TELL:
                            hearTell(sender, hear.message);
                            break;
                        default:
                            hear(sender, hear.message);
                            break;
                        }
                    }
                }
            } else {
                throw new Error("receives an illegal packet.");
            }
        }
    }

    /** This exception is thrown in order to exit from {@link #act()
     *  act} method.
     */
    protected static class ActionCommandException extends Exception {
        public ActionCommandException() { super(); }
    }

    protected void rest() throws ActionCommandException {
        socket().akRest(self().id);
        throw new ActionCommandException();
    }

    protected void say(String message) {
        if (m_numUtterance ++ < utteranceLimit())
            socket().akSay(self().id, message);
    }

    protected void tell(String message) {
        if (m_numUtterance ++ < utteranceLimit())
            socket().akTell(self().id, message);
    }
}
